//  (C) Copyright David Abrahams 2000. Permission to copy, use, modify, sell and
//  distribute this software is granted provided this copyright notice appears
//  in all copies. This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//
//  The author gratefully acknowleges the support of Dragon Systems, Inc., in
//  producing this work.

#ifndef EXTENSION_CLASS_DWA052000_H_
# define EXTENSION_CLASS_DWA052000_H_

# include "pyconfig.h"
# include "subclass.h"
# include <vector>
# include "constructors.h"
# include "none.h"
# include "objects.h"
# include "functions.h"
# include <memory>
# include "init_function.h"
# include <typeinfo>
# include <boost/smart_ptr.hpp>

namespace py {

// forward declarations
class ExtensionInstance;
template <class T> class InstanceHolder;
template <class T, class U> class InstanceValueHolder;
template <class Ptr, class T> class InstancePtrHolder;

MetaClass<ExtensionInstance>* extension_meta_class();
ExtensionInstance* get_extension_instance(PyObject* p);
void report_missing_instance_data(ExtensionInstance*, Class<ExtensionInstance>*, const std::type_info&);
void report_missing_ptr_data(ExtensionInstance*, Class<ExtensionInstance>*, const std::type_info&);
void report_missing_class_object(const std::type_info&);
void report_released_smart_pointer(const std::type_info&);
    
template <class T, class U>
struct ExtensionClassFromPython
{
};

template <class T>
T* check_non_null(T* p)
{
    if (p == 0)
        report_released_smart_pointer(typeid(T));
    return p;
}

template <class T> class HeldInstance;

#ifdef PY_NO_INLINE_FRIENDS_IN_NAMESPACE // back to global namespace for this GCC bug
}
#endif

template <class T>
class PyExtensionClassConvertersBase
{
 public:
    // Convert to T*
    friend T* from_python(PyObject* obj, py::Type<T*>)
    {
        // Downcast to an ExtensionInstance, then find the actual T
        py::ExtensionInstance* self = py::get_extension_instance(obj);
        typedef std::vector<py::InstanceHolderBase*>::const_iterator Iterator;
        for (Iterator p = self->wrapped_objects().begin();
             p != self->wrapped_objects().end(); ++p)
        {
            py::InstanceHolder<T>* held = dynamic_cast<py::InstanceHolder<T>*>(*p);
            if (held != 0)
                return held->target();
        }
        py::report_missing_instance_data(self, class_object, typeid(T));
        throw py::ErrorAlreadySet();
    }

    // Convert to PtrType, where PtrType can be dereferenced to obtain a T.
    template <class PtrType>
    static PtrType& ptr_from_python(PyObject* obj, py::Type<PtrType>)
    {
        // Downcast to an ExtensionInstance, then find the actual T
        py::ExtensionInstance* self = py::get_extension_instance(obj);
        typedef std::vector<py::InstanceHolderBase*>::const_iterator Iterator;
        for (Iterator p = self->wrapped_objects().begin();
             p != self->wrapped_objects().end(); ++p)
        {
            py::InstancePtrHolder<PtrType, T>* held =
                dynamic_cast<py::InstancePtrHolder<PtrType, T>*>(*p);
            if (held != 0)
                return held->ptr();
        }
        py::report_missing_ptr_data(self, class_object, typeid(T));
        throw py::ErrorAlreadySet();
    }

    template <class PtrType>
    static PyObject* ptr_to_python(PtrType x)
        {
            py::PyPtr<py::ExtensionInstance> result(create_instance(true));
            result->add_implementation(
                std::auto_ptr<py::InstanceHolderBase>(
                    new py::InstancePtrHolder<PtrType,T>(x)));
            return result.release();
        }

    static py::PyPtr<py::ExtensionInstance> create_instance(bool seek_base)
        {
            if (class_object == 0)
                py::report_missing_class_object(typeid(T));
            
            py::Class<py::ExtensionInstance>* class_
                = seek_base && class_object->bases().size() > 0
                ? py::Downcast<py::Class<py::ExtensionInstance> >(
                    class_object->bases()[0].get()).get()
                : class_object;
            
            return py::PyPtr<py::ExtensionInstance>(new py::ExtensionInstance(class_));
        }


    // Convert to const T*
    friend const T* from_python(PyObject* p, py::Type<const T*>)
        { return from_python(p, py::Type<T*>()); }

    // Convert to T&
    friend T& from_python(PyObject* p, py::Type<T&>)
        { return *py::check_non_null(from_python(p, py::Type<T*>())); }

    // Convert to const T&
    friend const T& from_python(PyObject* p, py::Type<const T&>)
        { return from_python(p, py::Type<T&>()); }

    // Convert to T
    friend const T& from_python(PyObject* p, py::Type<T>)
        { return from_python(p, py::Type<T&>()); }

    friend std::auto_ptr<T>& from_python(PyObject* p, py::Type<std::auto_ptr<T>&>)
        { return ptr_from_python(p, py::Type<std::auto_ptr<T> >()); }
    
    friend std::auto_ptr<T>& from_python(PyObject* p, py::Type<std::auto_ptr<T> >)
        { return ptr_from_python(p, py::Type<std::auto_ptr<T> >()); }
    
    friend const std::auto_ptr<T>& from_python(PyObject* p, py::Type<const std::auto_ptr<T>&>)
        { return ptr_from_python(p, py::Type<std::auto_ptr<T> >()); }

    friend PyObject* to_python(std::auto_ptr<T> x)
        { return ptr_to_python(x); }

    friend boost::shared_ptr<T>& from_python(PyObject* p, py::Type<boost::shared_ptr<T>&>)
        { return ptr_from_python(p, py::Type<boost::shared_ptr<T> >()); }
    
    friend boost::shared_ptr<T>& from_python(PyObject* p, py::Type<boost::shared_ptr<T> >)
        { return ptr_from_python(p, py::Type<boost::shared_ptr<T> >()); }
    
    friend const boost::shared_ptr<T>& from_python(PyObject* p, py::Type<const boost::shared_ptr<T>&>)
        { return ptr_from_python(p, py::Type<boost::shared_ptr<T> >()); }

    friend PyObject* to_python(boost::shared_ptr<T> x)
        { return ptr_to_python(x); }
    
 protected:
    static void register_class(py::Class<py::ExtensionInstance>*);
    static void unregister_class(py::Class<py::ExtensionInstance>*);
 private:
    static py::Class<py::ExtensionInstance>* class_object;
};

// This class' only job is to define from_python and to_python converters for T
// and U. T is the class the user really intends to wrap. U is a class derived
// from T with some virtual function overriding boilerplate, or if there are no
// virtual functions, U = HeldInstance<T>.
template <class T, class U = py::HeldInstance<T> >
class PyExtensionClassConverters : public PyExtensionClassConvertersBase<T>
{
    typedef PyExtensionClassConvertersBase<T> Base;
 public:
#ifdef BOOST_MSVC
    // Convert return values of type T to python objects.  What happens if T is
    // not copyable? Apparently there is no problem with g++ or MSVC unless this
    // is actually used. With a conforming compiler we will have a problem.
    friend PyObject* to_python(const T& x)
        {
            py::PyPtr<py::ExtensionInstance> result(create_instance(false));
            result->add_implementation(
                std::auto_ptr<py::InstanceHolderBase>(
                    new py::InstanceValueHolder<T,U>(result.get(), x)));
            return result.release();
        }
#else
    friend py::Type<U> py_holder_type(const T&)
        { return py::Type<U>(); }
#endif
};

#ifndef BOOST_MSVC
template <class T, class U>
py::InstanceHolderBase*
py_copy_to_new_value_holder(py::ExtensionInstance* p, const T& x, py::Type<U>)
{
    return new py::InstanceValueHolder<T,U>(p, x);
}

template <class T>
PyObject* to_python(const T& x)
{
    py::PyPtr<py::ExtensionInstance> result(
        PyExtensionClassConvertersBase<T>::create_instance(false));
    result->add_implementation(
        std::auto_ptr<py::InstanceHolderBase>(
            py_copy_to_new_value_holder(result.get(), x, py_holder_type(x))));
    return result.release();
}
#endif

#ifdef PY_NO_INLINE_FRIENDS_IN_NAMESPACE // back from global namespace for this GCC bug
namespace py {
using ::PyExtensionClassConverters;
#endif

template <class T> class InstanceHolder;

// An easy way to make an extension base class which wraps T. Note that Python
// subclasses of this class will simply be Class<ExtensionInstance> objects.
//
// U should be a class derived from T which overrides virtual functions with
// boilerplate code to call back into Python. See extclass_demo.h for examples.
// 
// U is optional, but you won't be able to override any member functions in
// Python which are called from C++ if you don't supply it. If you just want to
// be able to use T in python without overriding member functions, you can omit
// U.
template <class T, class U = HeldInstance<T> >
class ExtensionClass : public Class<ExtensionInstance>,
    PyExtensionClassConverters<T, U>, // This generates the to_python/from_python functions
    boost::noncopyable
{
 public:
    // GCC 2.95.2 choked on the use of typeid(T).name() as a default parameter:
    // "sorry, not implemented: `method_call_expr' not supported by dump_expr",
    // so we do the job inside the constructor instead
    ExtensionClass(const char* name = 0);
    ~ExtensionClass();

    // define constructors
    template <class A1, class A2, class A3, class A4, class A5>
    void def(Constructor<A1, A2, A3, A4, A5>)
#if 0
        { add_constructor(              // the following incantation builds a Signature1, 
            Signature0()                // Signature2,...Signature5 object for selecting a
            .prepend(Type<A5>::Id())    // constructor. It _should_ all get optimized away.
            .prepend(Type<A4>::Id())    // Just another workaround for the lack of partial
            .prepend(Type<A3>::Id())    // specialization in MSVC
            .prepend(Type<A2>::Id())
            .prepend(Type<A1>::Id()));
        }
#else
        { add_constructor(              // the following incantation builds a Signature1, 
            prepend(Type<A1>::Id(),    // constructor. It _should_ all get optimized away.
            prepend(Type<A2>::Id(),    // Just another workaround for the lack of partial
            prepend(Type<A3>::Id(),    // specialization in MSVC
            prepend(Type<A4>::Id(),
            prepend(Type<A5>::Id(),
                    Signature0()))))));
        }
#endif

    // define member functions. In fact this works for free functions, too -
    // they act like static member functions, or if they start with the
    // appropriate self argument (as a pointer), they can be used just like
    // ordinary member functions -- just like Python!
    template <class Fn>
    Ptr def(Fn fn, const char* name)
        {
            return add_method(new_member_function(fn), name);
        }

    template <class Fn, class DefaultFn>
    Ptr def(Fn fn, const char* name, DefaultFn default_fn)
        {
            add_default_method(new_member_function(default_fn), name);
            return add_method(new_member_function(fn), name);
        }
    
 private:
    Ptr add_method(Ptr method, const char* name);
    void add_default_method(Ptr method, const char* name);

    typedef InstanceValueHolder<T,U> Holder;
    
    template <class Signature>
    void add_constructor(Signature sig)
        { dict()[String("__init__").reference()] = InitFunction<Holder>::create(sig); }
};

#include "extclass_pygen.h"

template <class PtrType, class HeldType>
class InstancePtrHolder : public InstanceHolder<HeldType>
{
 public:
    HeldType* target() { return &*m_ptr; }
    PtrType& ptr() { return m_ptr; }

    InstancePtrHolder(PtrType ptr) : m_ptr(ptr) {}
 private:
    PtrType m_ptr;
};

class ExtensionInstance : public Instance
{
 public:
    ExtensionInstance(PyTypeObject* class_);
    ~ExtensionInstance();
    
    void add_implementation(std::auto_ptr<InstanceHolderBase> holder);

    typedef std::vector<InstanceHolderBase*> WrappedObjects;
    const WrappedObjects& wrapped_objects() const
        { return m_wrapped_objects; }
 private:
    WrappedObjects m_wrapped_objects;
};

//
// Template function implementations
//

template <class T, class U>
ExtensionClass<T, U>::ExtensionClass(const char* name)
    : Class<ExtensionInstance>(
        extension_meta_class(),
        String(name ? name : typeid(T).name()),
        Tuple(), Dict())
{
    register_class(this);
}

template <class T, class U>
ExtensionClass<T, U>::~ExtensionClass()
{
    unregister_class(this);
}

// Add the given callable object to this Python ExtensionClass as a method with
// the given name, returning a reference to the callable object
template <class T, class U>
Ptr ExtensionClass<T, U>::add_method(Ptr method, const char* name)
{
    // If we have created a special Python base class which wraps C++ classes
    // derived from T, the method should really be added there. Target will be
    // that Python class object.
    Class<ExtensionInstance>* target = (bases().size() == 0)
        ? this
        : Downcast<Class<ExtensionInstance> >(bases()[0].get()).get();

    // Add the attribute to the computed target
    target->setattr(name, method.get());

    // If it is a special member function it should be enabled both here and there.
    enable_named_method(target, name);
    enable_named_method(this, name);
    return method;
}

// Add the given callable object to this Python ExtensionClass as the
// overridable default implementation of a method with the given name which will
// be called from C++ through a virtual function.
template <class T, class U>
void ExtensionClass<T, U>::add_default_method(Ptr method, const char* name)
{
// A rather complicated thing is going on here in order to make a very specific
// class of cases work. When wrapping the following C++:
//
// struct Base {
//     Base();                  // will be constructed from Python
//     virtual int f() const    // might be called from C++
//         { return 1; }        // default implementation
// };
//
// struct Derived : Base {
//     int f() const { return 0; } // overridden in C++
// };
//
// boost::shared_ptr<Base> factory(bool selector) {
//     return boost::shared_ptr<Base>(selector ? new Base : new Derived);
// }
//
// Normally we would use the same Python ExtensionClass object to represent both
// Base and boost::shared_ptr<Base>, since they have essentially the same
// operations (see the comment on InstanceHolder in extclass_pygen.h for
// details). If there was no need to override Base::f() in Python, that would
// work fine. In this case, since f() is virtual, the programmer must provide a
// subclass of Base which calls back into Python:
//
// struct BaseCallback : Base {
//     BaseCallback(PyObject* self) : m_self(self) {}
//     int f() const { return py::Callback<int>::call_method(m_self, "f"); }
//     static int default_f(const Base* self) const { self->Base::f(); }
// };
//
// default_f() is what gets registered under the name "f" in the "Base"
// ExtensionClass' attribute dict. When C++ calls f() on a wrapped instance of
// Base, we call back into Python to find the "f" attribute, which calls
// default_f() (unless it has been overridden in Python) and in turn the default
// implementation (Base::f()) is called.
//
// Now consider what happens when the Python programmer writes
//    >>> factory(0).f()
//
// The shared_ptr<Derived> which is created on the C++ side is converted by
// to_python() into a Python instance of the "Base" ExtensionClass. Then Python
// looks up the "f" attribute, and finds the wrapper for default_f(), which it
// calls. That calls Base::f(), returning 1. What we really wanted was a call to
// Derived::f(), returning 0.
//
// In this case we actually need a different Python ExtensionClass to represent
// C++ subclasses of Base. When the first default method implementation is added
// to an ExtensionClass, we "push" all of the non-default methods up into a
// newly-created base class of the "Base" ExtensionClass, called
// "Base_base". "Base's" attribute dict contains only default method
// implementations.
// 
// A Python call to factory() then results in an object of class "Base_base",
// whose "f" method is bound to Base::f() - since this is a virtual function
// pointer, the member function actually called is determined by the
// most-derived class that implements f().
//
// A Python call to Base() results in an object of class "Base" wrapping a
// BaseCallback object, whose "f" method is bound to BaseCallback::default_f()
// ...which calls Base::f() explicitly.
    if (bases().size() == 0)
    {
        Class<ExtensionInstance>* new_base
            = new Class<ExtensionInstance>(
                extension_meta_class(), this->name() + String("_base"), Tuple(),
                dict());
        
        add_base(Ptr(as_object(new_base)));
        
        // We have transferred everything in our dict into the base class, so
        // clear our dict now. It will henceforth contain only default method
        // implementations.
        dict() = Dict();
    }
    dict()[String(name).reference()] = method;
}

#ifdef PY_NO_INLINE_FRIENDS_IN_NAMESPACE // Back to the global namespace for this GCC bug
}
#endif

template <class T>
void PyExtensionClassConvertersBase<T>::register_class(py::Class<py::ExtensionInstance>* p)
{
    // You're not expected to create more than one of these!
    assert(class_object == 0);
    class_object = p;
}

template <class T>
void PyExtensionClassConvertersBase<T>::unregister_class(py::Class<py::ExtensionInstance>* p)
{
    // The user should be destroying the same object they created.
    assert(class_object == p);
    (void)p; // unused in shipping version
    class_object = 0;
}

//
// Static data member declaration.
//
template <class T>
py::Class<py::ExtensionInstance>* PyExtensionClassConvertersBase<T>::class_object;

#ifdef PY_NO_INLINE_FRIENDS_IN_NAMESPACE // Back from the global namespace for this GCC bug
namespace py {
#endif


} // namespace py

#endif // EXTENSION_CLASS_DWA052000_H_
