//  (C) Copyright David Abrahams 2000. Permission to copy, use, modify, sell and
//  distribute this software is granted provided this copyright notice appears
//  in all copies. This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//
//  The author gratefully acknowleges the support of Dragon Systems, Inc., in
//  producing this work.

#include "newtypes.h"
#include "pyptr.h" // for handle_exception()
#include "module.h"
#include "none.h"
#include <set>

namespace py {

// TODO: is there a problem with calling convention here, or can I really pass a
// pointer to a C++ linkage function as a C-linkage function pointer? The
// compilers seem to swallow it, but is it legal? Symantec C++ for Mac didn't
// behave this way, FWIW.
// Using C++ linkage allows us to keep the virtual function members of
// TypeObjectBase private and use friendship to get them called.

//extern "C" {
//static
PyObject* do_instance_call(PyObject* instance, PyObject* args, PyObject* keywords)
{
    try
    {
        return static_cast<TypeObjectBase*>(instance->ob_type)->instance_call(instance, args, keywords);
    }
    catch(...)
    {
        handle_exception();
        return 0;
    }
}

//static
void do_instance_dealloc(PyObject* instance)
{
    try
    {
        static_cast<TypeObjectBase*>(instance->ob_type)->instance_dealloc(instance);
    }
    catch(...)
    {
        assert(!"exception during destruction!");
        handle_exception();
    }
}

//static
PyObject* do_instance_getattr(PyObject* instance, char* name)
{
    try
    {
        return static_cast<TypeObjectBase*>(instance->ob_type)->instance_getattr(instance, name);
    }
    catch(...)
    {
        handle_exception();
        return 0;
    }
}

//static
int do_instance_setattr(PyObject* instance, char* name, PyObject* value)
{
    try
    {
        return static_cast<TypeObjectBase*>(instance->ob_type)->instance_setattr(instance, name, value);
    }
    catch(...)
    {
        handle_exception();
        return -1;
    }
}

//static
int do_instance_mapping_length(PyObject* instance)
{
    try
    {
        const int outcome =
            static_cast<TypeObjectBase*>(instance->ob_type)->instance_mapping_length(instance);
        
        if (outcome < 0)
        {
            PyErr_SetString(PyExc_ValueError, "__len__() should return >= 0");
            return -1;
        }
        return outcome;
    }
    catch(...)
    {
        handle_exception();
        return -1;
    }
}

//static
PyObject* do_instance_mapping_subscript(PyObject* instance, PyObject* index)
{
    try
    {
        return static_cast<TypeObjectBase*>(instance->ob_type)->instance_mapping_subscript(instance, index);
    }
    catch(...)
    {
        handle_exception();
        return 0;
    }
}

//static
int do_instance_mapping_ass_subscript(PyObject* instance, PyObject* index, PyObject* value)
{
    try
    {
        return static_cast<TypeObjectBase*>(instance->ob_type)->instance_mapping_ass_subscript(instance, index, value);
    }
    catch(...)
    {
        handle_exception();
        return -1;
    }
}

//}

template <class Cmp1>
struct lexicographic1
{
    bool operator()(typename Cmp1::first_argument_type x1, typename Cmp1::second_argument_type y1) {
        return Cmp1()(x1, y1);
    }
};

template <class Cmp1, class Cmp2>
struct lexicographic2
{
    bool operator()(typename Cmp1::first_argument_type x1, typename Cmp2::first_argument_type x2,
                    typename Cmp1::second_argument_type y1, typename Cmp2::second_argument_type y2) {
        return Cmp1()(x1, y1) ? true
            : Cmp1()(y1, x1) ? false
            : Cmp2()(x2, y2);
    }
};

template <class Cmp1, class Cmp2, class Cmp3>
struct lexicographic3
{
    bool operator()(typename Cmp1::first_argument_type x1, typename Cmp2::first_argument_type x2, typename Cmp3::first_argument_type x3,
                    typename Cmp1::second_argument_type y1, typename Cmp2::second_argument_type y2, typename Cmp3::second_argument_type y3) {
        return Cmp1()(x1, y1) ? true
            : Cmp1()(y1, x1) ? false
            : Cmp2()(x2, y2) ? true
            : Cmp2()(y2, x2) ? false
            : Cmp3()(x3, y3);
    }
};

struct MappingMethods : PyMappingMethods
{
    MappingMethods() { mp_length = 0; mp_subscript = 0; mp_ass_subscript = 0; }
    MappingMethods(const PyMappingMethods& rhs) : PyMappingMethods(rhs) {}

    template <class MemberPtr, class Fn>
    static MappingMethods* enable(const PyMappingMethods* base, MemberPtr p, Fn f) {
#if !defined(__MWERKS__) || __MWERKS__ > 0x6000 // CWPro6 bug
        MappingMethods new_value(base != 0 ? *base : MappingMethods());
#else
        MappingMethods new_value;
        if (base != 0)
        {
            new_value = *base;
        }
#endif
        new_value.*p = f;
        return const_cast<MappingMethods*>(&*mapping_methods.insert(new_value).first);
    }
    
    bool operator<(const MappingMethods& rhs) const {
        return lexicographic3<std::less<inquiry>,std::less<binaryfunc>,std::less<objobjargproc> >()
            (mp_length, mp_subscript, mp_ass_subscript, rhs.mp_length, rhs.mp_subscript, rhs.mp_ass_subscript);
    }
 private:
    struct Set;
    static Set mapping_methods;
};

struct MappingMethods::Set : std::set<MappingMethods> {};
MappingMethods::Set MappingMethods::mapping_methods;

//std::set<MappingMethods> MappingMethods::mapping_methods;

void TypeObjectBase::enable_instance_call()
{
    this->tp_call = do_instance_call;
}

void TypeObjectBase::enable_instance_getattr()
{
    this->tp_getattr = do_instance_getattr;
}

void TypeObjectBase::enable_instance_setattr()
{
    this->tp_setattr = do_instance_setattr;
}

void TypeObjectBase::enable_instance_mapping_length()
{
    this->tp_as_mapping = MappingMethods::enable(
        this->tp_as_mapping, &PyMappingMethods::mp_length, do_instance_mapping_length);
}

void TypeObjectBase::enable_instance_mapping_subscript()
{
    this->tp_as_mapping = MappingMethods::enable(
        this->tp_as_mapping, &PyMappingMethods::mp_subscript, do_instance_mapping_subscript);
}

void TypeObjectBase::enable_instance_mapping_ass_subscript()
{
    this->tp_as_mapping = MappingMethods::enable(
        this->tp_as_mapping, &PyMappingMethods::mp_ass_subscript, do_instance_mapping_ass_subscript);
}

TypeObjectBase::TypeObjectBase(PyTypeObject* t)
    : PythonType(t)
{
    this->tp_dealloc = do_instance_dealloc;
}

PyObject* TypeObjectBase::instance_call(PyObject* /*instance*/, PyObject* /*args*/, PyObject* /*kw*/) const
{
    PyErr_SetString(PyExc_RuntimeError, "Unimplemented instance_call");
    return none();
}

PyObject* TypeObjectBase::instance_getattr(PyObject* /*instance*/, const char* /*name*/) const
{
    PyErr_SetString(PyExc_RuntimeError, "Unimplemented instance_getattr");
    return none();
}

int TypeObjectBase::instance_setattr(PyObject* /*instance*/, const char* /*name*/, PyObject* /*value*/) const
{
    PyErr_SetString(PyExc_RuntimeError, "Unimplemented instance_setattr");
    return -1;
}

int TypeObjectBase::instance_mapping_length(PyObject*) const
{
    PyErr_SetString(PyExc_RuntimeError, "Unimplemented instance_mapping_length");
    return -1;
}

PyObject* TypeObjectBase::instance_mapping_subscript(PyObject*, PyObject*) const
{
    PyErr_SetString(PyExc_RuntimeError, "Unimplemented instance_mapping_subscript");
    return 0;
}

int TypeObjectBase::instance_mapping_ass_subscript(PyObject*, PyObject*, PyObject*) const
{
    PyErr_SetString(PyExc_RuntimeError, "Unimplemented instance_mapping_ass_subscript");
    return -1;
}

TypeObjectBase::~TypeObjectBase()
{
}

void TypeObjectBase::replicate_enabled_methods(const TypeObjectBase* other)
{
    PyTypeObject* me = this;
    const PyTypeObject* him = other;

    // OK to use memcpy here because a PyTypeObject is POD.
    memcpy(&me->tp_dealloc, &him->tp_dealloc,
           reinterpret_cast<char*>(&me->tp_flags) - reinterpret_cast<char*>(&me->tp_dealloc));
}

}
