//  (C) Copyright David Abrahams 2000. Permission to copy, use, modify, sell and
//  distribute this software is granted provided this copyright notice appears
//  in all copies. This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//
//  The author gratefully acknowleges the support of Dragon Systems, Inc., in
//  producing this work.

#ifndef TYPES_DWA051800_H_
# define TYPES_DWA051800_H_

//  Usage:
//  class X : public
//          py::Callable<
//          py::Getattrable <
//          py::Setattrable<PythonObject, X> > >
//  {
//  public:
//      Ptr call(args, kw);
//      Ptr getattr(args, kw);
//      Ptr setattr(args, kw);
//  };

# include "pyconfig.h"
# include "base_object.h"
# include "cast.h"
# include <typeinfo>

namespace py {

class TypeObjectBase : public PythonType
{
 public:
    explicit TypeObjectBase(PyTypeObject* type_type);
    virtual ~TypeObjectBase();

    // Make the enabled methods of this type match those of other
    void replicate_enabled_methods(const TypeObjectBase* other);

 public: // Functions to fill in the function slots in the underlying PyTypeObject.
    // Python uses some of them, like tp_call, to determine properties of the
    // type.  For others, such as tp_setattr, there are useful default behaviors
    // when the function pointer is 0, so we don't want to fill them in prematurely.
    void enable_instance_call();
    void enable_instance_getattr();
    void enable_instance_setattr();
    void enable_instance_mapping_length();
    void enable_instance_mapping_subscript();
    void enable_instance_mapping_ass_subscript();

 private: // Callbacks for basic type functionality.
    // Inactive unless the corresponding enable_xxxx has been called.
    virtual PyObject* instance_call(PyObject* instance, PyObject* args, PyObject* kw) const;
    virtual PyObject* instance_getattr(PyObject* instance, const char* name) const;
    virtual int instance_setattr(PyObject* instance, const char* name, PyObject* value) const;

    // Dealloc is a special case, since every type needs a nonzero tp_dealloc slot.
    virtual void instance_dealloc(PyObject*) const = 0;

 private: // Callbacks for mapping methods
    virtual int instance_mapping_length(PyObject*) const;
    virtual PyObject* instance_mapping_subscript(PyObject*, PyObject*) const ;
    virtual int instance_mapping_ass_subscript(PyObject*, PyObject*, PyObject*) const;
    
 private: // friend functions which occupy the actual tp_xxxx slots
    friend PyObject* do_instance_call(PyObject*, PyObject*, PyObject*);
    friend void do_instance_dealloc(PyObject* instance);
    friend PyObject* do_instance_getattr(PyObject* instance, char* name);
    friend int do_instance_setattr(PyObject* instance, char* name, PyObject* value);

 private: // friend functions for mapping types
    friend int do_instance_mapping_length(PyObject*);
    friend PyObject* do_instance_mapping_subscript(PyObject*, PyObject*);
    friend int do_instance_mapping_ass_subscript(PyObject*, PyObject*, PyObject*);
};

template <class T>
class TypeObject : public TypeObjectBase
{
 public:
    typedef T Instance;
    
    TypeObject(PyTypeObject* type_type, const char* name = 0)
        : TypeObjectBase(type_type)
        {
            this->tp_name = const_cast<char*>(name ? name : typeid(Instance).name());
        }

 private: // Overridable behaviors.
    // Called when the reference count goes to zero. The default implementation
    // is "delete p". If you have not allocated your object with operator new or
    // you have other constraints, you'll need to override this
    virtual void dealloc(T* p) const;
    
 private: // Implementation of TypeObjectBase hooks. Do not reimplement in derived classes.
    void instance_dealloc(PyObject*) const;
};

//
// Type objects
//
template <class Base>
class Callable : public Base
{
 public:
    typedef Callable Properties; // Convenience for derived class construction
    typedef typename Base::Instance Instance;
    Callable(PyTypeObject* type_type, const char* name = 0);
 private:
    PyObject* instance_call(PyObject* instance, PyObject* args, PyObject* kw) const;
};

template <class Base>
class Getattrable : public Base
{
 public:
    typedef Getattrable Properties; // Convenience for derived class construction
    typedef typename Base::Instance Instance;
    Getattrable(PyTypeObject* type_type, const char* name = 0);
 private:
    PyObject* instance_getattr(PyObject* instance, const char* name) const;
};

template <class Base>
class Setattrable : public Base
{
 public:
    typedef Setattrable Properties; // Convenience for derived class construction
    typedef typename Base::Instance Instance;
    Setattrable(PyTypeObject* type_type, const char* name = 0);
 private:
    int instance_setattr(PyObject* instance, const char* name, PyObject* value) const;
};

//
// Member function definitions
//

// TypeObject<>
template <class T>
void TypeObject<T>::instance_dealloc(PyObject* instance) const
{
    this->dealloc(Downcast<Instance>(instance).get());
}

template <class T>
void TypeObject<T>::dealloc(T* instance) const
{
    delete instance;
}

// Callable
template <class Base>
Callable<Base>::Callable(PyTypeObject* type_type, const char* name)
    : Base(type_type, name)
{
    this->enable_instance_call();
}
        
template <class Base>
PyObject* Callable<Base>::instance_call(PyObject* instance, PyObject* args, PyObject* kw) const
{
    return Downcast<Instance>(instance)->call(args, kw);
}

// Getattrable
template <class Base>
Getattrable<Base>::Getattrable(PyTypeObject* type_type, const char* name)
    : Base(type_type, name)
{
    this->enable_instance_getattr();
}
        
template <class Base>
PyObject* Getattrable<Base>::instance_getattr(PyObject* instance, const char* name) const
{
    return Downcast<Instance>(instance)->getattr(name);
}

// Setattrable
template <class Base>
Setattrable<Base>::Setattrable(PyTypeObject* type_type, const char* name)
    : Base(type_type, name)
{
    this->enable_instance_setattr();
}
        
template <class Base>
int Setattrable<Base>::instance_setattr(PyObject* instance, const char* name, PyObject* value) const
{
    return Downcast<Instance>(instance)->setattr(name, value);
}

}

#endif // TYPES_DWA051800_H_
