//  (C) Copyright David Abrahams 2000. Permission to copy, use, modify, sell and
//  distribute this software is granted provided this copyright notice appears
//  in all copies. This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//
//  The author gratefully acknowleges the support of Dragon Systems, Inc., in
//  producing this work.

// TODO: Move inline implementations from objects.cpp here

#include "objects.h"
#include "none.h"

namespace py {

Object::Object(Ptr p)
    : m_p(p) {}
    
// Return a reference to the held object
Ptr Object::reference() const
{
    return m_p;
}

// Return a raw pointer to the held object
PyObject* Object::get() const
{
    return m_p.get();
}

Tuple::Tuple(std::size_t n)
    : Object(Ptr(PyTuple_New(n)))
{
    for (std::size_t i = 0; i < n; ++i)
        PyTuple_SET_ITEM(get(), i, none());
}
    
Tuple::Tuple(Ptr p)
    : Object(p)
{
    assert(accepts(p));
}

PyTypeObject* Tuple::type_object()
{
    return &PyTuple_Type;
}

bool Tuple::accepts(Ptr p)
{
    return PyTuple_Check(p.get());
}

std::size_t Tuple::size() const
{
    return PyTuple_Size(get());
}

Ptr Tuple::operator[](std::size_t pos) const
{
    return Ptr(PyTuple_GetItem(get(), static_cast<int>(pos)),
               Ptr::new_ref);
}

void Tuple::set_item(std::size_t pos, const Ptr& rhs)
{
    int failed = PyTuple_SetItem(
        get(), static_cast<int>(pos), Ptr(rhs).release()); // A reference is stolen here.
    assert(failed == 0);
}

Tuple Tuple::slice(int low, int high) const
{
    return Tuple(Ptr(PyTuple_GetSlice(get(), low, high), Ptr::new_ref));
}

Tuple& Tuple::operator+=(const Tuple& rhs)
{
    return *this = *this + rhs;
}


// Construct from an owned PyObject*.
// Precondition: p must point to a python string.
String::String(Ptr p)
    : Object(p) { assert(accepts(p)); }
    
String::String(const char* s)
    : Object(Ptr(PyString_FromString(s))) {}

String::String(const char* s, std::size_t length)
    : Object(Ptr(PyString_FromStringAndSize(s, length))) {}

String::String(const char* s, Interned)
    : Object(Ptr(PyString_InternFromString(s))) {}

#if 0
String::String(const char* s, std::size_t length, Interned)
    : Object(Ptr(PyString_InternFromStringAndSize(s, length))) {}
#endif

String::String(const String& rhs)
    : Object(rhs.reference()) {}

// Get the type object for Strings
PyTypeObject* String::type_object()
{ return &PyString_Type; }

// Return true if the given object is a python String
bool String::accepts(Ptr o)
{ return PyString_Check(o.get()); }

// Return the length of the string.
std::size_t String::size() const
{
    int size = PyString_GET_SIZE(get());
    assert(size >= 0);
    return static_cast<std::size_t>(size);
}

// Returns a null-terminated representation of the contents of string.
// The pointer refers to the internal buffer of string, not a copy.
// The data must not be modified in any way. It must not be de-allocated. 
const char* String::c_str() const
{ return PyString_AS_STRING(get()); }

String& String::operator+=(const String& rhs)
{ return *this = *this + rhs; }

void String::intern()
{ // UNTESTED!!
    *this = String(Ptr(PyString_InternFromString(c_str()), Ptr::borrowed));
}

Dict::Dict(Ptr p)
    : Object(p) { assert(accepts(p)); }

Dict::Dict()
    : Object(Ptr(PyDict_New())) {}

PyTypeObject* Dict::type_object()
{ return &PyDict_Type; }

bool Dict::accepts(Ptr p)
{ return PyDict_Check(p.get()); }

void Dict::clear()
{ PyDict_Clear(get()); }

const Ptr& Dict::Proxy::operator=(const Ptr& rhs) {
    if (PyDict_SetItem(m_dict.get(), m_key.get(), rhs.get()) == -1)
        throw ErrorAlreadySet();
    return rhs;
}

Dict::Proxy::operator Ptr() const
{ return Ptr(m_dict->ob_type->tp_as_mapping->mp_subscript(m_dict.get(), m_key.get()), Ptr::borrowed); }

Dict::Proxy::Proxy(const Ptr& dict, const Ptr& key)
    : m_dict(dict), m_key(key) {}

Dict::Proxy Dict::operator[](Ptr key)
{ return Proxy(reference(), key); }
    
Ptr Dict::operator[](Ptr key) const {
    // An odd MSVC bug causes the ".operator Ptr()" to be needed
    return Proxy(reference(), key).operator Ptr();
}

    
Ptr Dict::get_item(const Ptr& key, const Ptr& _default /* = Ptr() */)
{
    PyObject* value_or_null = PyDict_GetItem(get(), key.get());
    if (value_or_null == 0 && !PyErr_Occurred())
        return _default;
    else
        return Ptr(value_or_null, Ptr::borrowed); // Will throw if there was another error
}
        
void Dict::erase(Ptr key) {
    if (PyDict_DelItem(get(), key.get()) == -1)
        throw ErrorAlreadySet();
}

Ptr Dict::items() const { return Ptr(PyDict_Items(get())); }
Ptr Dict::keys() const { return Ptr(PyDict_Keys(get())); }
Ptr Dict::values() const { return Ptr(PyDict_Values(get())); }

std::size_t Dict::size() const { return static_cast<std::size_t>(PyDict_Size(get())); }

Dict::Proxy Dict::operator[](const Object& key)
{
    return this->operator[](key.reference());
}

Ptr Dict::operator[](const Object& key) const
{
    return this->operator[](key.reference());
}

Ptr Dict::get_item(const Object& key, Ptr default_)
{
    return this->get_item(key.reference(), default_);
}
        
void Dict::erase(const Object& key)
{
    this->erase(key.reference());
}


// TODO: iterator support

String operator+(const String& x, const String& y)
{
    PyObject* io_string = x.reference().release();
    PyString_Concat(&io_string, y.get());
    return String(Ptr(io_string));
}

String operator%(const String& format, const Tuple& args)
{
    return String(Ptr(PyString_Format(format.get(), args.reference().get())));
}

Tuple operator+(const Tuple& x, const Tuple& y)
{
    Tuple result(x.size() + y.size());
    for (std::size_t xi = 0; xi < x.size(); ++xi)
        result.set_item(xi, x[xi]);
    for (std::size_t yi = 0; yi < y.size(); ++yi)
        result.set_item(yi + x.size(), y[yi]);
    return result;
}

}
