//  (C) Copyright David Abrahams 2000. Permission to copy, use, modify, sell and
//  distribute this software is granted provided this copyright notice appears
//  in all copies. This software is provided "as is" without express or implied
//  warranty, and with no claim as to its suitability for any purpose.
//
//  The author gratefully acknowleges the support of Dragon Systems, Inc., in
//  producing this work.

#include "subclass.h"
#include "functions.h"
#include "singleton.h"
#include <cstddef>
#include "callback.h"

namespace py {

Instance::Instance(PyTypeObject* class_)
    : PythonObject(class_)
{
}

PyObject* Instance::getattr(const char* name, bool use_special_function)
{
    Ptr local_attribute = m_name_space.get_item(String(name).reference());
    
    if (local_attribute.get())
        return local_attribute.release();

    // Check its class. 
    PyObject* function =
        PyObject_GetAttrString(as_object(this->ob_type), const_cast<char*>(name));

    Ptr class_attribute;
    if (!use_special_function || function != 0)
    {
        // This will throw if the attribute wasn't found
        class_attribute = Ptr(function);
    }
    else
    {
        // Suspend the error while we try the __getattr__ method (if any).
        SuspendError suspended_error(SuspendError::discard_new_error);
        PyObject* const getattr_method = PyObject_GetAttrString(
            as_object(this->ob_type), const_cast<char*>("__getattr__"));
        // If there is no such method, throw now.
        suspended_error.throw_if_error();
        
        Ptr owner(getattr_method);
        
        // Call it to get the attribute.
        return PyEval_CallFunction(getattr_method, const_cast<char*>("(OO)"), this, to_python(name));
    }

    if (!PyCallable_Check(class_attribute.get()))
    {
        PyErr_Clear();
        return class_attribute.release();
    }
    else
    {
        return new BoundFunction(Ptr(this, Ptr::borrowed), class_attribute);
    }
}

int Instance::setattr(const char* name, PyObject* value)
{
    // See if there's an appropriate "special" function (if value == 0, we are
    // to delete an attribute).
    PyObject* f = PyObject_GetAttrString(
        as_object(this->ob_type), const_cast<char*>(value ? "__setattr__" : "__delattr__"));

    if (f != 0)
    {
        Ptr owner(f); // take possession of the reference

        // Now call the "special" function that was found
        expect_non_null(
            value != 0
            ? PyEval_CallFunction(f, const_cast<char*>("(OOO)"),
                                  this, to_python(name), value)
            : PyEval_CallFunction(f, const_cast<char*>("(OO)"),
                                  this, to_python(name))
            );
        return 0;
    }
    
    PyErr_Clear(); // Nothing was found; clear the python error state
    
    if (value == 0) // Try to remove the attribute from our name space
    {
        const int result = PyDict_DelItemString(
            m_name_space.reference().get(), const_cast<char*>(name));
        
        if (result < 0)
        {
            PyErr_Clear();
            PyErr_SetString(PyExc_AttributeError, "delete non-existing instance attribute");
        }
        return result;
    }
    else // Change the specified item in our name space
    {
        return PyDict_SetItemString(
            m_name_space.reference().get(), const_cast<char*>(name), value);
    }
}

PyObject* Instance::call(PyObject* args, PyObject* keywords)
{
    return PyEval_CallObjectWithKeywords(
        Ptr(getattr("__call__")).get(), // take possession of the result from getattr()
        args, keywords);
}

int Instance::length()
{
    return Callback<int>::call_method(this, "__len__");
}

PyObject* Instance::get_subscript(PyObject* key)
{
    return Callback<PyObject*>::call_method(this, "__getitem__", key);
}

void Instance::set_subscript(PyObject* key, PyObject* value)
{
    if (value == 0)
        Callback<void>::call_method(this, "__delitem__", key);
    else
        Callback<void>::call_method(this, "__setitem__", key, value);
}

// Enable any special methods which are enabled in the base class.
void enable_special_base_methods(TypeObjectBase* derived, const Tuple& bases)
{
    for (std::size_t i = 0; i < bases.size(); ++i)
    {
        PyTypeObject* base = Downcast<PyTypeObject>(bases[i].get());
        
        if (base->tp_call)
            derived->enable_instance_call();
        if (base->tp_getattr)
            derived->enable_instance_getattr();
        if (base->tp_setattr)
            derived->enable_instance_setattr();
        
        if (base->tp_as_mapping)
        {
            if (base->tp_as_mapping->mp_length)
                derived->enable_instance_mapping_length();
            if (base->tp_as_mapping->mp_subscript)
                derived->enable_instance_mapping_subscript();
            if (base->tp_as_mapping->mp_ass_subscript)
                derived->enable_instance_mapping_ass_subscript();
        }
    }
}

// Enable the special handler for methods of the given name, if any.
void enable_named_method(TypeObjectBase* type_object, const char* name)
{
    struct Enablers
    {
        const char* name;
        void (TypeObjectBase::*enabler)();
    };

    static Enablers enablers[] =
    {
        { "__call__", &TypeObjectBase::enable_instance_call },
        { "__getattr__", &TypeObjectBase::enable_instance_getattr },
        { "__setattr__", &TypeObjectBase::enable_instance_setattr },
        { "__len__", &TypeObjectBase::enable_instance_mapping_length },
        { "__getitem__", &TypeObjectBase::enable_instance_mapping_subscript },
        { "__setitem__", &TypeObjectBase::enable_instance_mapping_ass_subscript },
        { "__delitem__", &TypeObjectBase::enable_instance_mapping_ass_subscript }
    };

    const std::size_t num_enablers = sizeof(enablers) / sizeof(enablers[0]);
    for (std::size_t i = 0; i < num_enablers; ++i)
    {
        if (!strcmp(name, enablers[i].name))
        {
            (type_object->*enablers[i].enabler)();
            break;
        }
    }
}

void add_current_module_name(Dict& name_space)
{
    static String module_key("__module__", String::interned);
    static String name_key("__name__", String::interned);
    
    Ptr existing_value = name_space.get_item(module_key);
    if (existing_value.get() == 0)
    {
        PyObject* globals = PyEval_GetGlobals();
        if (globals != 0) // Why don't we throw in this case? Who knows? This is
        {                 // what Python does for class objects!
            PyObject* module_name = PyDict_GetItem(globals, name_key.get());
            if (module_name != 0)
            {
                name_space[module_key] = Ptr(module_name, Ptr::borrowed);
            }
        }
    }
}

} // namespace py
